@@ -6,32 +6,62 @@
  */
 #include "punctuation.h"
 #include "notifications_public.h"
+#include <algorithm>
 #include <boost/iostreams/device/file_descriptor.hpp>
 #include <boost/iostreams/stream_buffer.hpp>
+#include <cstdint>
+#include <exception>
 #include <fcitx-config/iniparser.h>
+#include <fcitx-config/rawconfig.h>
+#include <fcitx-utils/capabilityflags.h>
 #include <fcitx-utils/charutils.h>
+#include <fcitx-utils/fs.h>
+#include <fcitx-utils/i18n.h>
+#include <fcitx-utils/key.h>
+#include <fcitx-utils/keysymgen.h>
 #include <fcitx-utils/log.h>
+#include <fcitx-utils/macros.h>
+#include <fcitx-utils/misc.h>
 #include <fcitx-utils/standardpath.h>
+#include <fcitx-utils/stringutils.h>
 #include <fcitx-utils/utf8.h>
+#include <fcitx/addoninstance.h>
+#include <fcitx/event.h>
 #include <fcitx/inputcontext.h>
 #include <fcitx/inputcontextmanager.h>
 #include <fcitx/inputmethodentry.h>
 #include <fcitx/inputmethodmanager.h>
+#include <fcitx/instance.h>
 #include <fcitx/statusarea.h>
 #include <fcitx/userinterfacemanager.h>
 #include <fcntl.h>
+#include <fstream>
+#include <ios>
+#include <istream>
+#include <iterator>
+#include <string>
 #include <string_view>
-#include <unordered_set>
+#include <unordered_map>
+#include <utility>
+#include <vector>
 
 using namespace fcitx;
 
 namespace {
 const std::string emptyString;
 const std::pair<std::string, std::string> emptyStringPair;
-} // namespace
 
 bool dontConvertWhenEn(uint32_t c) { return c == '.' || c == ','; }
 
+std::string langByPath(const std::string &path) {
+    constexpr std::string_view prefix = "punctuationmap/";
+    if (stringutils::startsWith(path, prefix)) {
+        return path.substr(prefix.size());
+    }
+    return "";
+}
+} // namespace
+
 class PunctuationState : public InputContextProperty {
 public:
     std::unordered_map<uint32_t, std::string> lastPuncStack_;
@@ -43,16 +73,44 @@ public:
     uint32_t notConvertedBackup_ = 0;
 };
 
-PunctuationProfile::PunctuationProfile(std::istream &in) {
+void PunctuationProfile::loadSystem(std::istream &in) {
+    load(in);
+    punctuationMapConfig_.syncDefaultValueToCurrent();
+}
+
+void PunctuationProfile::resetDefaultValue() {
+    punctuationMapConfig_ = PunctuationMapConfig();
+    punctuationMapConfig_.syncDefaultValueToCurrent();
+}
+
+void PunctuationProfile::addEntry(uint32_t key, const std::string &value,
+                                  const std::string &value2) {
+    decltype(puncMap_)::mapped_type::value_type p;
+    p.first = value;
+    p.second = value2;
+    puncMap_[key].push_back(p);
+
+    std::string punc = utf8::UCS4ToUTF8(key);
+    auto *configValue = punctuationMapConfig_.entries.mutableValue();
+    configValue->emplace_back();
+    PunctuationMapEntryConfig &entryConfig = configValue->back();
+    entryConfig.key.setValue(punc);
+    entryConfig.mapResult1.setValue(value);
+    entryConfig.mapResult2.setValue(value2);
+}
+
+void PunctuationProfile::load(std::istream &in) {
+    puncMap_.clear();
+    auto *configValue = punctuationMapConfig_.entries.mutableValue();
+    configValue->clear();
+
     std::string strBuf;
     while (std::getline(in, strBuf)) {
-        auto pair = stringutils::trimInplace(strBuf);
-        std::string::size_type start = pair.first, end = pair.second;
-        if (start == end) {
+        std::string_view trimmed = stringutils::trimView(strBuf);
+        if (trimmed.empty()) {
             continue;
         }
-        std::string text(strBuf.begin() + start, strBuf.begin() + end);
-        auto tokens = stringutils::split(strBuf, FCITX_WHITESPACE);
+        auto tokens = stringutils::split(trimmed, FCITX_WHITESPACE);
         if (tokens.size() != 2 && tokens.size() != 3) {
             continue;
         }
@@ -67,22 +125,82 @@ PunctuationProfile::PunctuationProfile(std::istream &in) {
             continue;
         }
         auto c = utf8::getChar(tokens[0]);
-        decltype(puncMap_)::mapped_type p;
-        p.first = tokens[1];
-        if (tokens.size() > 2) {
-            p.second = tokens[2];
-        }
-        puncMap_.emplace(c, std::move(p));
+        addEntry(c, tokens[1], tokens.size() > 2 ? tokens[2] : "");
     }
 }
 
+void PunctuationProfile::set(const RawConfig &config) {
+    PunctuationMapConfig newConfig;
+    newConfig.load(config);
+
+    puncMap_.clear();
+    auto *configValue = punctuationMapConfig_.entries.mutableValue();
+    configValue->clear();
+
+    const auto &entries = *newConfig.entries;
+    // remove duplicate entry
+    for (const auto &entry : entries) {
+        if (entry.key->empty() || entry.mapResult1->empty()) {
+            continue;
+        }
+        if (utf8::lengthValidated(*entry.key) != 1) {
+            continue;
+        }
+        auto c = utf8::getChar(*entry.key);
+
+        addEntry(c, *entry.mapResult1, *entry.mapResult2);
+    }
+}
+
+void PunctuationProfile::save(std::string_view name) const {
+    StandardPath::global().safeSave(
+        StandardPath::Type::PkgData,
+        stringutils::concat("punctuation/", profilePrefix, name),
+        [this](int fd) {
+            for (const auto &entry : *punctuationMapConfig_.entries) {
+                fs::safeWrite(fd, entry.key->data(), entry.key->size());
+                fs::safeWrite(fd, " ", sizeof(char));
+                fs::safeWrite(fd, entry.mapResult1->data(),
+                              entry.mapResult1->size());
+                if (!entry.mapResult2->empty()) {
+                    fs::safeWrite(fd, " ", sizeof(char));
+                    fs::safeWrite(fd, entry.mapResult2->data(),
+                                  entry.mapResult2->size());
+                }
+                fs::safeWrite(fd, "\n", sizeof(char));
+            }
+            return true;
+        });
+}
+
 const std::pair<std::string, std::string> &
 PunctuationProfile::getPunctuation(uint32_t unicode) const {
     auto iter = puncMap_.find(unicode);
     if (iter == puncMap_.end()) {
         return emptyStringPair;
     }
-    return iter->second;
+    return iter->second[0];
+}
+
+std::vector<std::string>
+PunctuationProfile::getPunctuations(uint32_t unicode) const {
+    auto iter = puncMap_.find(unicode);
+    if (iter == puncMap_.end()) {
+        return {};
+    }
+    // Return only first if the result size is 1.
+    // This allows single paired symbol to work.
+    if (iter->second.size() == 1) {
+        return {iter->second[0].first};
+    }
+    std::vector<std::string> result;
+    for (const auto &punc : iter->second) {
+        result.push_back(punc.first);
+        if (!punc.second.empty()) {
+            result.push_back(punc.second);
+        }
+    }
+    return result;
 }
 
 Punctuation::Punctuation(Instance *instance)
@@ -109,15 +227,17 @@ Punctuation::Punctuation(Instance *instance)
                 state->lastIsEngOrDigit_ = '\0';
             }
         });
-    auto processKeyEvent = [this](const KeyEventBase &event) {
-        const auto &keyEvent = static_cast<const ForwardKeyEvent &>(event);
+    auto processKeyEvent = [this](const KeyEventBase &keyEvent) {
         auto *state = keyEvent.inputContext()->propertyFor(&factory_);
         if (keyEvent.isRelease()) {
             return;
         }
-        if (!event.accepted()) {
+        if (!keyEvent.accepted()) {
             if (keyEvent.key().isUAZ() || keyEvent.key().isLAZ() ||
-                keyEvent.key().isDigit()) {
+                keyEvent.key().isDigit() ||
+                (keyEvent.key().sym() >= FcitxKey_KP_0 &&
+                 keyEvent.key().sym() <= FcitxKey_KP_9 &&
+                 !keyEvent.key().hasModifier())) {
                 state->lastIsEngOrDigit_ =
                     Key::keySymToUnicode(keyEvent.key().sym());
             } else {
@@ -234,14 +354,9 @@ Punctuation::Punctuation(Instance *instance)
             // Scan through the surrounding text
             if (!state->lastPuncStackBackup_.empty() &&
                 state->lastPuncStack_.empty()) {
-                auto range =
-                    utf8::MakeUTF8CharRange(MakeIterRange(text.begin(), end));
-                for (auto iter = std::begin(range); iter != std::end(range);
-                     iter++) {
-                    auto charRange = iter.charRange();
-                    std::string_view chr(
-                        &*charRange.first,
-                        std::distance(charRange.first, charRange.second));
+                auto range = utf8::MakeUTF8StringViewRange(
+                    MakeIterRange(text.begin(), end));
+                for (std::string_view chr : range) {
                     auto puncIter = std::find_if(
                         state->lastPuncStackBackup_.begin(),
                         state->lastPuncStackBackup_.end(),
@@ -259,40 +374,55 @@ Punctuation::~Punctuation() {}
 
 void Punctuation::reloadConfig() {
     readAsIni(config_, "conf/punctuation.conf");
-    populateConfig();
+    loadProfiles();
 }
 
-void Punctuation::populateConfig() {
-    auto files = StandardPath::global().multiOpen(StandardPath::Type::PkgData,
-                                                  "punctuation", O_RDONLY,
-                                                  filter::Prefix("punc.mb."));
+void Punctuation::loadProfiles() {
+    auto systemFiles = StandardPath::global().locate(
+        StandardPath::Type::PkgData, "punctuation",
+        filter::Prefix(std::string(PunctuationProfile::profilePrefix)),
+        filter::Not(filter::User()));
+    auto allFiles = StandardPath::global().locate(
+        StandardPath::Type::PkgData, "punctuation",
+        filter::Prefix(std::string(PunctuationProfile::profilePrefix)));
+
+    // Remove non-exist profiles.
     auto iter = profiles_.begin();
     while (iter != profiles_.end()) {
-        if (!files.count("punc.mb." + iter->first)) {
+        if (!allFiles.count(stringutils::concat(
+                PunctuationProfile::profilePrefix.size(), iter->first))) {
             iter = profiles_.erase(iter);
         } else {
             iter++;
         }
     }
 
-    for (const auto &file : files) {
-        if (file.first.size() <= 8) {
+    for (const auto &file : allFiles) {
+        if (file.first.size() <= PunctuationProfile::profilePrefix.size()) {
             continue;
         }
-        auto lang = file.first.substr(8);
+        auto lang = file.first.substr(PunctuationProfile::profilePrefix.size());
 
+        auto iter = systemFiles.find(file.first);
+        const bool hasSystemFile = iter != systemFiles.end();
+        bool hasUserFile = true;
+        if (hasSystemFile && iter->second == file.second) {
+            hasUserFile = false;
+        }
         try {
-            boost::iostreams::stream_buffer<
-                boost::iostreams::file_descriptor_source>
-                buffer(file.second.fd(),
-                       boost::iostreams::file_descriptor_flags::
-                           never_close_handle);
-            std::istream in(&buffer);
-            PunctuationProfile newProfile(in);
-            profiles_[lang] = std::move(newProfile);
+            if (hasSystemFile) {
+                std::ifstream in(iter->second, std::ios::in | std::ios::binary);
+                profiles_[lang].loadSystem(in);
+            } else {
+                profiles_[lang].resetDefaultValue();
+            }
+            if (hasUserFile) {
+                std::ifstream in(file.second, std::ios::in | std::ios::binary);
+                profiles_[lang].load(in);
+            }
         } catch (const std::exception &e) {
-            FCITX_LOG(Warn)
-                << "Error when load profile " << file.first << ": " << e.what();
+            FCITX_WARN() << "Error when load profile " << file.first << ": "
+                         << e.what();
         }
     }
 }
@@ -381,6 +511,19 @@ const std::string &Punctuation::cancelLast(const std::string &language,
     return emptyString;
 }
 
+std::vector<std::string>
+Punctuation::getPunctuationCandidates(const std::string &language,
+                                      uint32_t unicode) {
+    if (!enabled()) {
+        return {};
+    }
+    auto iter = profiles_.find(language);
+    if (iter == profiles_.end()) {
+        return {};
+    }
+    return getPunctuations(language, unicode);
+}
+
 const std::pair<std::string, std::string> &
 Punctuation::getPunctuation(const std::string &language, uint32_t unicode) {
     if (!*config_.enabled) {
@@ -395,4 +538,41 @@ Punctuation::getPunctuation(const std::string &language, uint32_t unicode) {
     return iter->second.getPunctuation(unicode);
 }
 
+std::vector<std::string>
+Punctuation::getPunctuations(const std::string &language, uint32_t unicode) {
+    if (!*config_.enabled) {
+        return {};
+    }
+
+    auto iter = profiles_.find(language);
+    if (iter == profiles_.end()) {
+        return {};
+    }
+
+    return iter->second.getPunctuations(unicode);
+}
+
+const fcitx::Configuration *
+Punctuation::getSubConfig(const std::string &path) const {
+    auto lang = langByPath(path);
+    if (lang.empty()) {
+        return nullptr;
+    }
+    if (auto iter = profiles_.find(lang); iter != profiles_.end()) {
+        return &iter->second.config();
+    }
+    return nullptr;
+}
+
+void Punctuation::setSubConfig(const std::string &path,
+                               const fcitx::RawConfig &config) {
+    std::string lang = langByPath(path);
+    auto iter = profiles_.find(lang);
+    if (iter == profiles_.end()) {
+        return;
+    }
+    iter->second.set(config);
+    iter->second.save(lang);
+}
+
 FCITX_ADDON_FACTORY(PunctuationFactory);
